/*
 *            August 1995, by Akira KITAUCHI
 *
 *              E-mail address: kitauchi@i.h.kyoto-u.ac.jp
 *              WWW home page : http://www.i.h.kyoto-u.ac.jp/~kitauchi/
 */
/*
   #include <stdio.h>
   #include <memory.h>
 */
typedef long int count_int;

#ifdef __STDC__
static void expand_PIC (char *, char *, int, int, int);
static void putword (int, char *);
static int compress (int, char *, char *, int);
static void output (int);
static void cl_block (void);
static void cl_hash (count_int);
static void char_init (void);
static void char_out (int);
static void flush_char (void);
#else
static void expand_PIC (), putword (), output (), cl_block (), cl_hash ();
static int compress ();
static void char_init (), char_out (), flush_char ();
#endif

/***********************************************************************/

PIC_to_GIF (picbuf, gifbuf, chN, chW, chH, fgcolor)
     char *picbuf, *gifbuf;
     int chN, chW, chH;
     int fgcolor;
{
  static char transGifStr[] =
  {
    0x21, 0xf9, 0x04, 0x01, 0x00, 0x00, 0x00, 0x00,
  };
  static int bgcolor = 0xffffff;
  char expbuf[32 * 34 * (GIFWIDTH + 1)], *dgif = gifbuf;
  int LeftOfs = 0, TopOfs = 0;
  int InitCodeSize = 2, Background = 0, BitsPerPixel = 1;
  int Interlace = 0;
  int i;

  strcpy (dgif, "GIF89a");
  dgif += 6;			/* the GIF magic number */

  putword (chW * chN, dgif);
  dgif += 2;			/* screen descriptor */
  putword (chH, dgif);
  dgif += 2;

  i = 0x80;			/* Yes, there is a color map */
  i |= (8 - 1) << 4;		/* OR in the color resolution (hardwired 8) */
  i |= (BitsPerPixel - 1);	/* OR in the # of bits per pixel */

  *dgif++ = i;
  *dgif++ = Background;		/* background color */
  *dgif++ = 0;			/* future expansion byte */

  *dgif++ = bgcolor >> 16;
  *dgif++ = bgcolor >> 8;
  *dgif++ = bgcolor;
  *dgif++ = fgcolor >> 16;
  *dgif++ = fgcolor >> 8;
  *dgif++ = fgcolor;

  memcpy (dgif, transGifStr, 8);
  dgif += 8;

  *dgif++ = ',';		/* image separator */

  /* Write the Image header */
  putword (LeftOfs, dgif);
  dgif += 2;
  putword (TopOfs, dgif);
  dgif += 2;
  putword (chW * chN, dgif);
  dgif += 2;
  putword (chH, dgif);
  dgif += 2;

  *dgif++ = Interlace ? 0x40 : 0x00;	/* Use Global Colormap, maybe Interlace */

  *dgif++ = InitCodeSize;

  expand_PIC (picbuf, expbuf, chN, chW, chH);
  dgif += compress (InitCodeSize + 1, expbuf, dgif, chN * chW * chH);

  *dgif++ = 0;			/* Write out a Zero-length packet (EOF) */
  *dgif++ = ';';		/* Write GIF file terminator */

  return (int) (dgif - gifbuf);
}

/******************************/
static void
expand_PIC (picbuf, expbuf, chN, chW, chH)
     char *picbuf, *expbuf;
     int chN, chW, chH;
{
  int bytes, lackb, data, h, n, b, i;

  bytes = (chW + 7) / 8;
  lackb = (8 * 100 - chW) % 8;

  memset (expbuf, 0, chN * chW * chH);

  for (h = 0; h < chH; h++)
    {
      memset (expbuf, 0, 8);
      for (n = 0; n < chN; n++)
	{
	  for (b = 0; b < bytes; b++)
	    {
	      data = *picbuf++;
	      for (i = 0; i < 8; i++)
		{
		  *expbuf++ |= data & 1;
		  data >>= 1;
		}
	    }
	  expbuf -= lackb;
	}
    }
}

/******************************/
static void
putword (w, dst)
     int w;
     char *dst;
{
  /* writes a 16-bit integer in GIF order (LSB first) */
  dst[0] = w;
  dst[1] = w >> 8;
}

/***********************************************************************/


static unsigned long cur_accum = 0;
static int cur_bits = 0;




#define min(a,b)        ((a>b) ? b : a)

#define XV_BITS	12		/* BITS was already defined on some systems */
#define MSDOS	1

#define HSIZE  5003		/* 80% occupancy */

typedef unsigned char char_type;


static int n_bits;		/* number of bits/code */
static int maxbits = XV_BITS;	/* user settable max # bits/code */
static int maxcode;		/* maximum code, given n_bits */
static int maxmaxcode = 1 << XV_BITS;	/* NEVER generate this */

#define MAXCODE(n_bits)     ( (1 << (n_bits)) - 1)

static count_int htab[HSIZE];
static unsigned short codetab[HSIZE];
#define HashTabOf(i)   htab[i]
#define CodeTabOf(i)   codetab[i]

static int hsize = HSIZE;	/* for dynamic table sizing */

/*
 * To save much memory, we overlay the table used by compress() with those
 * used by decompress().  The tab_prefix table is the same size and type
 * as the codetab.  The tab_suffix table needs 2**BITS characters.  We
 * get this from the beginning of htab.  The output stack uses the rest
 * of htab, and contains characters.  There is plenty of room for any
 * possible stack (stack used to be 8000 characters).
 */

#define tab_prefixof(i) CodeTabOf(i)
#define tab_suffixof(i)        ((char_type *)(htab))[i]
#define de_stack               ((char_type *)&tab_suffixof(1<<XV_BITS))

static int free_ent = 0;	/* first unused entry */

/*
 * block compression parameters -- after all codes are used up,
 * and compression rate changes, start over.
 */
static int clear_flg = 0;

static long int in_count = 1;	/* length of input */
static long int out_count = 0;	/* # of codes output (for debugging) */

/*
 * compress stdin to stdout
 *
 * Algorithm:  use open addressing double hashing (no chaining) on the 
 * prefix code / next character combination.  We do a variant of Knuth's
 * algorithm D (vol. 3, sec. 6.4) along with G. Knott's relatively-prime
 * secondary probe.  Here, the modular division first probe is gives way
 * to a faster exclusive-or manipulation.  Also do block compression with
 * an adaptive reset, whereby the code table is cleared when the compression
 * ratio decreases, but after the table fills.  The variable-length output
 * codes are re-sized at this point, and a special CLEAR code is generated
 * for the decompressor.  Late addition:  construct the table according to
 * file size for noticeable speed improvement on small files.  Please direct
 * questions about this implementation to ames!jaw.
 */

static int g_init_bits;
static char *g_outbuf;

static int ClearCode;
static int EOFCode;


/********************************************************/
static int
compress (init_bits, indata, outdata, len)
     int init_bits;
     char *indata, *outdata;
     int len;
{
  register long fcode;
  register int i = 0;
  register int c;
  register int ent;
  register int disp;
  register int hsize_reg;
  register int hshift;

  /*
   * Set up the globals:  g_init_bits - initial number of bits
   *                      g_outbuf    - pointer to output buffer
   */
  g_init_bits = init_bits;
  g_outbuf = outdata;

  /* initialize 'compress' globals */
  maxbits = XV_BITS;
  maxmaxcode = 1 << XV_BITS;
  memset ((char *) htab, 0, sizeof (htab));
  memset ((char *) codetab, 0, sizeof (codetab));
  hsize = HSIZE;
  free_ent = 0;
  clear_flg = 0;
  in_count = 1;
  out_count = 0;
  cur_accum = 0;
  cur_bits = 0;


  /*
   * Set up the necessary values
   */
  out_count = 0;
  clear_flg = 0;
  in_count = 1;
  maxcode = MAXCODE (n_bits = g_init_bits);

  ClearCode = (1 << (init_bits - 1));
  EOFCode = ClearCode + 1;
  free_ent = ClearCode + 2;

  char_init ();
  ent = *indata++;
  len--;

  hshift = 0;
  for (fcode = (long) hsize; fcode < 65536L; fcode *= 2L)
    hshift++;
  hshift = 8 - hshift;		/* set hash code range bound */

  hsize_reg = hsize;
  cl_hash ((count_int) hsize_reg);	/* clear hash table */

  output (ClearCode);

  while (len)
    {
      c = *indata++;
      len--;
      in_count++;

      fcode = (long) (((long) c << maxbits) + ent);
      i = (((int) c << hshift) ^ ent);	/* xor hashing */

      if (HashTabOf (i) == fcode)
	{
	  ent = CodeTabOf (i);
	  continue;
	}

      else if ((long) HashTabOf (i) < 0)	/* empty slot */
	goto nomatch;

      disp = hsize_reg - i;	/* secondary hash (after G. Knott) */
      if (i == 0)
	disp = 1;

    probe:
      if ((i -= disp) < 0)
	i += hsize_reg;

      if (HashTabOf (i) == fcode)
	{
	  ent = CodeTabOf (i);
	  continue;
	}

      if ((long) HashTabOf (i) > 0)
	goto probe;

    nomatch:
      output (ent);
      out_count++;
      ent = c;

      if (free_ent < maxmaxcode)
	{
	  CodeTabOf (i) = free_ent++;	/* code -> hashtable */
	  HashTabOf (i) = fcode;
	}
      else
	cl_block ();
    }

  /* Put out the final code */
  output (ent);
  out_count++;
  output (EOFCode);

  return (int) (g_outbuf - outdata);
}


/*****************************************************************
 * TAG( output )
 *
 * Output the given code.
 * Inputs:
 *      code:   A n_bits-bit integer.  If == -1, then EOF.  This assumes
 *              that n_bits =< (long)wordsize - 1.
 * Outputs:
 *      Outputs code to the file.
 * Assumptions:
 *      Chars are 8 bits long.
 * Algorithm:
 *      Maintain a BITS character long buffer (so that 8 codes will
 * fit in it exactly).  Use the VAX insv instruction to insert each
 * code in turn.  When the buffer fills up empty it and start over.
 */

static
unsigned long masks[] =
{0x0000, 0x0001, 0x0003, 0x0007, 0x000F,
 0x001F, 0x003F, 0x007F, 0x00FF,
 0x01FF, 0x03FF, 0x07FF, 0x0FFF,
 0x1FFF, 0x3FFF, 0x7FFF, 0xFFFF};

static void
output (code)
     int code;
{
  cur_accum &= masks[cur_bits];

  if (cur_bits > 0)
    cur_accum |= ((long) code << cur_bits);
  else
    cur_accum = code;

  cur_bits += n_bits;

  while (cur_bits >= 8)
    {
      char_out ((unsigned int) (cur_accum & 0xff));
      cur_accum >>= 8;
      cur_bits -= 8;
    }

  /*
   * If the next entry is going to be too big for the code size,
   * then increase it, if possible.
   */

  if (free_ent > maxcode || clear_flg)
    {

      if (clear_flg)
	{
	  maxcode = MAXCODE (n_bits = g_init_bits);
	  clear_flg = 0;
	}
      else
	{
	  n_bits++;
	  if (n_bits == maxbits)
	    maxcode = maxmaxcode;
	  else
	    maxcode = MAXCODE (n_bits);
	}
    }

  if (code == EOFCode)
    {
      /* At EOF, write the rest of the buffer */
      while (cur_bits > 0)
	{
	  char_out ((unsigned int) (cur_accum & 0xff));
	  cur_accum >>= 8;
	  cur_bits -= 8;
	}

      flush_char ();
    }
}


/********************************/
static void
cl_block ()			/* table clear for block compress */
{
  /* Clear out the hash table */

  cl_hash ((count_int) hsize);
  free_ent = ClearCode + 2;
  clear_flg = 1;

  output (ClearCode);
}


/********************************/
static void
cl_hash (hsize)			/* reset code table */
     register count_int hsize;
{
  register count_int *htab_p = htab + hsize;
  register long i;
  register long m1 = -1;

  i = hsize - 16;
  do
    {				/* might use Sys V memset(3) here */
      *(htab_p - 16) = m1;
      *(htab_p - 15) = m1;
      *(htab_p - 14) = m1;
      *(htab_p - 13) = m1;
      *(htab_p - 12) = m1;
      *(htab_p - 11) = m1;
      *(htab_p - 10) = m1;
      *(htab_p - 9) = m1;
      *(htab_p - 8) = m1;
      *(htab_p - 7) = m1;
      *(htab_p - 6) = m1;
      *(htab_p - 5) = m1;
      *(htab_p - 4) = m1;
      *(htab_p - 3) = m1;
      *(htab_p - 2) = m1;
      *(htab_p - 1) = m1;
      htab_p -= 16;
    }
  while ((i -= 16) >= 0);

  for (i += 16; i > 0; i--)
    *--htab_p = m1;
}

/******************************************************************************
 *
 * GIF Specific routines
 *
 ******************************************************************************/

/*
 * Number of characters so far in this 'packet'
 */
static int a_count;

/*
 * Set up the 'byte output' routine
 */
static void
char_init ()
{
  a_count = 0;
}

/*
 * Define the storage for the packet accumulator
 */
static char accum[256];

/*
 * Add a character to the end of the current packet, and if it is 254
 * characters, flush the packet to disk.
 */
static void
char_out (c)
     int c;
{
  accum[a_count++] = c;
  if (a_count >= 254)
    flush_char ();
}

/*
 * Flush the packet to disk, and reset the accumulator
 */
static void
flush_char ()
{
  if (a_count > 0)
    {
      *g_outbuf++ = a_count;
      memcpy (g_outbuf, accum, a_count);
      g_outbuf += a_count;
      a_count = 0;
    }
}
